//////////////////////////////////////////////////
// JIST (Java In Simulation Time) Project
// Timestamp: <CBR.java Tue 2004/04/06 11:57:22 barr pompom.cs.cornell.edu>
//

// Copyright (C) 2004 by Cornell University
// All rights reserved.
// Refer to LICENSE for terms and conditions of use.

package driver;

import jist.swans.field.Field;
import jist.swans.field.Mobility;
import jist.swans.field.Placement;
import jist.swans.radio.RadioNoise;
import jist.swans.radio.RadioNoiseIndep;
import jist.swans.radio.RadioInfo;
import jist.swans.mac.MacAddress;
import jist.swans.mac.MacDumb;
import jist.swans.net.NetAddress;
import jist.swans.net.NetIp;
import jist.swans.net.PacketLoss;
import jist.swans.trans.TransUdp;
import jist.swans.trans.TransInterface;
import jist.swans.route.RouteDhg;
import jist.swans.route.RouteInterface;
import jist.swans.route.RouteDsr;
import jist.swans.route.RouteAodv;
import jist.swans.route.RouteZrp;
import jist.swans.route.RouteZrpNdp;
import jist.swans.route.RouteZrpIarp;
import jist.swans.route.RouteZrpBrp;
import jist.swans.route.RouteZrpIerp;
import jist.swans.route.RouteZrpZdp;
import jist.swans.misc.Util;
import jist.swans.misc.Mapper;
import jist.swans.misc.Location;
import jist.swans.misc.Message;
import jist.swans.misc.MessageBytes;
import jist.swans.Constants;

import jist.runtime.JistAPI;

import jargs.gnu.*;

import java.util.*;

import org.apache.log4j.Logger;


public class MyCBR
{
  /** Default port number to send and receive packets. */
  private static final int PORT = 3001;
  public static final long SEND_JITTER       = 10 * Constants.MILLI_SECOND;
  

  //////////////////////////////////////////////////
  // locals
  //

  /** The number of clients currently transmitting. */
  private static int numClientsTransmitting = 0;
  
  /** MyCBR logger. */
  private static Logger log = Logger.getLogger(MyCBR.class.getName());  


  //////////////////////////////////////////////////
  // server
  //

  /**
   * The interface for server nodes in the simulation.
   */
  public interface ServerInterface extends JistAPI.Proxiable
  {
    /** Starts the server. */
    void run();
  }

  /**
   * A server node in the simulation.  This node simply listens for incoming
   * packets and prints a message for each one that it receives.
   */
  public static class Server implements ServerInterface
  {
    /** The proxy interface for this object. */
    private ServerInterface self;
    /** The UDP interface used by this node. */
    private TransInterface.TransUdpInterface udp;
    /** The IP address of this node. */
    private NetAddress localAddr;
    /** The number of packets this node has received. */
    private int packetsReceived;
    
    private int id;
    
    private long packetsSizeReceived;
    private long timeFirstPacket;
    private long timeLastPacket;
    private double sommeDelai=0.0f;
    private double delaiMax=0.0f;
    
    private ArrayList idPaquetReceived;
    /**
     * Creates a new <code>Server</code>.
     *
     * @param udp the UDP interface used by this node
     * @param localAddr the IP address of this node
     */
    public Server(TransInterface.TransUdpInterface udp, NetAddress localAddr,int id)
    {
      this.self = (ServerInterface)JistAPI.proxy(this, ServerInterface.class);
      this.udp = udp;
      this.localAddr = localAddr;
      this.packetsReceived = 0;
      
      this.packetsSizeReceived=0;
      this.timeFirstPacket=0;
      this.timeLastPacket=0;
      this.sommeDelai=0.0f;
      this.delaiMax=0.0f; 
      this.id=id;
      
      idPaquetReceived = new ArrayList();
    }
    
    public Server(TransInterface.TransUdpInterface udp, NetAddress localAddr)
    {
    	this(udp,localAddr,0);
    }
    

    /**
     * Returns the proxy interface for this object.
     *
     * @return the proxy interface for this object.
     */
    public ServerInterface getProxy()
    {
      return self;
    }

    /** {@inheritDoc} */
    public void run()
    {
      // Create a UDP event handler that just makes a note of incoming packets
      TransInterface.SocketHandler handler = new TransInterface.SocketHandler()
        {
          public void receive(Message msg, NetAddress src, int srcPort)
          {

        	
        	  
        	
        	byte bMsg[] = new byte[msg.getSize()];
        	msg.getBytes(bMsg, 0);
        	String strMsg=new String(bMsg);
        	String[] strs=strMsg.split(" ");
        	int timesend=(Integer.decode(strs[2]).intValue());
        	Integer num=Integer.decode(strs[0]);
        	int delay =  (int)((float)JistAPI.getTime()/(float)Constants.MILLI_SECOND) -  timesend ;
        	
        	if(idPaquetReceived.indexOf(num)==-1)
        		idPaquetReceived.add(num);
        	else{
        		//System.out.println("duplicate paquet "+num);
        		return;
        	}
      
        	
        	sommeDelai+=delay;
        	if(delay>delaiMax)
        		delaiMax=delay;
        	
        	if(packetsReceived==0)
        		timeFirstPacket=timesend;  
        	
        	timeLastPacket=timesend;
        	
        	if(log.isInfoEnabled())
        		log.info(JistAPI.getTime()+" Received_message " + num + " "+id);
            
        	packetsReceived++;
            packetsSizeReceived+=msg.getSize();
            
            
            
          }
        };

      udp.addSocketHandler(PORT, handler);
    }

	public double getDelayMax() {
		return delaiMax;
	}

	public NetAddress getLocalAddr() {
		return localAddr;
	}

	public int getPacketsReceived() {
		return packetsReceived;
	}
	
	public long getPacketsSizeReceived() {
		return packetsSizeReceived;
	}

	public double getDelayAvg(){
		if(packetsReceived==0)
				return 0;
		
		return (double)sommeDelai/(double)packetsReceived;
	}
	
	
	
  }

  //////////////////////////////////////////////////
  // client
  //

  /**
   * The interface for client nodes in the simulation.
   */
  public interface ClientInterface extends JistAPI.Proxiable
  {
    /** Starts the client. */
    void run();

    /**
     * Sends message number <code>i</code> to this client's corresponding
     * server.
     *
     * @param i the number of the message to send
     */
    void sendMessage(int i);
  }

  /**
   * A client node in the simulation.  Each client node sends
   * <code>numTransmissions</code> packets to its corresponding server at a
   * rate of one per second, then terminates.
   */
  public static class Client implements ClientInterface
  {
    /** self-referencing proxy entity. */
    private ClientInterface self;
    /** UDP entity used by this node. */
    private TransInterface.TransUdpInterface udp;
    /** local IP address. */
    private NetAddress localAddr;
    /** address of the server that corresponds to this client. */
    private NetAddress serverAddr;
    /** number of outgoing transmissions. */
    private int transmissions;
    
    private int sizePacket;
    private long pause;
    private int packetsSended;
    private int id;
    
    private boolean stop=false;
    


	/**
     * Creates a new <code>Client</code>.
     *
     * @param udp the UDP interface for this node to use
     * @param bitrate number of bits by second
     * @param sizePacket packet size in byte
     * @param timeSimu send time
     * @param localAddr this node's IP address
     * @param serverAddr the IP address of the server to send messages to
     */
    public Client(TransInterface.TransUdpInterface udp, float bitrate,
    	int sizePacket,long timeSimu,NetAddress localAddr, NetAddress serverAddr,int id)
    {
      self = (ClientInterface)JistAPI.proxy(this, ClientInterface.class);
      this.udp = udp;
      this.localAddr = localAddr;
      this.serverAddr = serverAddr;
      this.transmissions = (int)((bitrate/((float)sizePacket*8.0f)) * ((float)timeSimu)) ;
      this.pause = (long)  ((((float)sizePacket*8.0f)/bitrate)*(float)Constants.SECOND);
      this.sizePacket = sizePacket;
      this.packetsSended=0;
      this.id=id;
    }
    
    public Client(TransInterface.TransUdpInterface udp, float bitrate,
        	int sizePacket,long timeSimu,NetAddress localAddr, NetAddress serverAddr)
        {
    		this(udp,bitrate,sizePacket,timeSimu,localAddr,serverAddr,0);
        }

    /**
     * Returns this object's proxy interface.
     *
     * @return this object's proxy interface
     */
    public ClientInterface getProxy()
    {
      return self;
    }
    
    public void stop(){
    	stop=true;
    }

    /** {@inheritDoc} */
    public void run()
    {

      // Send the appropriate number of messages to the corresponding server
      for (int i=0; i<transmissions; i++)
      {
    	if(stop)break;
        JistAPI.sleep(pause + (long)(Math.random() * SEND_JITTER));
        packetsSended++;
        self.sendMessage(i+1);
      }
    }

    /** {@inheritDoc} */
    public void sendMessage(int i)
    {
    	long t=JistAPI.getTime()/Constants.MILLI_SECOND;
      
		String msgStr=new String(i+" hi "+t);
		
		//bourage
		int nb=sizePacket-msgStr.length();
		for(int j=0;j<nb;j++)
			msgStr+=" ";      

    	if(log.isInfoEnabled())
    		log.info(JistAPI.getTime()+" Send message " + i+" hi "+t);
		
		  MessageBytes msg = new MessageBytes(msgStr);
		  udp.send(msg, serverAddr, PORT, PORT, Constants.NET_PRIORITY_NORMAL);
		  if(i==transmissions)
		  {
		    numClientsTransmitting--;
		    // System.out.println("Memory: " + jist.runtime.Util.getUsedMemory());
		  }
    }
    

    public NetAddress getLocalAddr() {
		return localAddr;
	}

	public NetAddress getServerAddr() {
		return serverAddr;
	}

	public int getPacketsSended() {
		return packetsSended;
	}    
    
  }

  /**
   * Starts the CBR simulation.
   *
   * @param args command-line arguments that may determine the parameters
   *   of the simulation
   */
  public static void main(String[] args)
  {
  }

} // class: CBR

