package jist.swans.route;

import jist.runtime.JistAPI;
import jist.swans.Constants;
import jist.swans.mac.MacAddress;
import jist.swans.misc.Message;
import jist.swans.misc.Util;
import jist.swans.net.NetAddress;
import jist.swans.net.NetInterface;
import jist.swans.net.NetMessage;
import jist.swans.net.NetMessage.Ip;


public class Cluster implements RouteInterface.Cluster {

	  /**
	   * Represents a HELLO message.
	   */
	  private static class HelloMessage implements Message
	  {
	    /** Size of HELLO Message in bytes. */
	    private static final int MESSAGE_SIZE = 13;
	    
	    /** Net address of node issuing HELLO message. */
	    private NetAddress ip;
	    
	    /** Sequence number of node issuing HELLO message. */
	    private int seqNum;
	    
	    /**Flags : |UNDECIDED|MEMBER|CH|GATEWAY|REPLY 
	     * 5 bits -> 1 octets
	     * */
	    private byte stat;
	    private boolean reply;
	    
	    private NetAddress clusterHead;

	    
	    /**
	     * Constructs new HELLO Message object.
	     * 
	     * @param ip net address of this node
	     * @param seqNum sequence number of this node
	     */
	    public HelloMessage(NetAddress ip, NetAddress clusterHead,int seqNum,byte stat,boolean reply)
	    {
	      this.ip = ip;
	      this.clusterHead = clusterHead;
	      this.seqNum = seqNum;
	      this.stat = stat;
	      this.reply=reply;
	    }
	    /**
	     * Returns HELLO message ip field.
	     * 
	     * @return Hello message ip field
	     */
	    public NetAddress getIp()
	    {
	      return ip;
	    }
	    /**
	     * Return size of packet.
	     * 
	     * @return size of packet
	     */
	    public int getSize()
	    {
	      return MESSAGE_SIZE;
	    }
	    /**
	     * Store packet into byte array.
	     *
	     * @param msg destination byte array
	     * @param offset byte array starting offset
	     */
	    public void getBytes(byte[] msg, int offset)
	    {
	      throw new RuntimeException("RouteReplyMessage.getBytes() not implemented.");    
	    }
		public byte getStat() {
			return stat;
		}
		public NetAddress getClusterHead() {
			return clusterHead;
		}    
	  }
	
	
	
	  /** The maximum amount of jitter before sending a packet. */
	  public static final long TRANSMISSION_JITTER = 1 * Constants.MILLI_SECOND;	

	  /** Timer delai. */
	  public static final long TIMER_JITER = 1 * Constants.SECOND;	
	  public static final long TIMER = 1 * Constants.SECOND;	
	  
	  
	  /** The maximum Time-To-Live for a MFR packet. */
	  public static final byte MAX_TTL             = (byte)127;
	  
	  /** Timeout member cluster */
	  public static final long TIMEOUT_MEMBER_CLUSTER = 2*TIMER;

	  
	  /**  Stats*/
	  public static final byte STAT_UNDECIDED          = (byte)0;	  
	  public static final byte STAT_MEMBER             = (byte)1;
	  public static final byte STAT_CLUSTERHEAD        = (byte)2;	  
	  public static final byte STAT_GATEWAY            = (byte)3;	  
	  
	  
	  
	  
	  
	  /** Network entity. */
	  private NetInterface netEntity;
	  /** Self-referencing proxy entity. */
	  private RouteInterface.Cluster self;

	  /** local network address. */
	  private NetAddress netAddr;
	  
	  private NetAddress clusterHead=null;
	  private byte stat = STAT_UNDECIDED;
	  private boolean receiveHelloCH=false;  
	  private  long lastReceiveHelloCH=0;

	  
	  
	public void log(String msg){
		System.out.println(JistAPI.getTime()/Constants.MILLI_SECOND + " " + netAddr + " "+msg);
	}	  
	
	
	public Cluster(NetAddress netAddr) {
		super();

	    	this.self = (RouteInterface.Cluster)JistAPI.proxy(this, RouteInterface.Cluster.class);
		this.netAddr = netAddr;
		
		setStat(STAT_UNDECIDED);
		sendHelloMessage();
		self.startTimer(((long)(Math.random() * TIMER_JITER)) + TIMER);
		
	}
	
	/**
	 * registre a node to clusterHead
	 * @param clusterHead
	 */
	public void registreToClusterHead(NetAddress clusterHead){
		this.clusterHead=clusterHead;
		log("Register to clusterhead "+clusterHead);
		setStat(STAT_MEMBER);
	}
	

	
	public void startTimer(long delay){
		JistAPI.sleep(delay);
		self.timeOut();
	}


	public void timeOut() {
		//log("timeout begin");
		// TODO Auto-generated method stub
		
		log("timeout");		
		if(!receiveHelloCH){
			receiveHelloCH=false;
			
			
			setStat(STAT_CLUSTERHEAD);
			sendHelloMessage();
			
			
			
		}
		self.startTimer(TIMER);
	}
	
	private void receiveHelloMessage(HelloMessage helloMsg){
		log("Receive a hello message: "+helloMsg.getIp());
		
		if ( helloMsg.getStat() == STAT_CLUSTERHEAD ){
			receiveHelloCH=true;
			lastReceiveHelloCH=JistAPI.getTime();
		}
		
		switch(helloMsg.getStat()){
		
			case STAT_UNDECIDED:
				break;
				
			//sender is a clusterhead
			case STAT_CLUSTERHEAD:
				
				switch ( stat ){
				
				case STAT_UNDECIDED:
					registreToClusterHead(helloMsg.getIp());
					//now i member ! (no break)
				case STAT_MEMBER:	
				case STAT_GATEWAY:
					/*if the last message from clusterhead is old
					 * register to a new clusterhead
					 */
					if((JistAPI.getTime()-lastReceiveHelloCH) > TIMEOUT_MEMBER_CLUSTER) {
						registreToClusterHead(helloMsg.getIp());
					}
					
					//the sender is not in clusterhead
					if(helloMsg.getIp()!=clusterHead)
						setStat(STAT_GATEWAY);

					
					sendReplyHelloMessage();
					
					break;					
				}
				
				
			break;
//			end sender is a clusterhead
			
			
			case STAT_GATEWAY:
			case STAT_MEMBER:
				switch ( stat ){
				
				case STAT_UNDECIDED:
					break;

				case STAT_CLUSTERHEAD:
					addMember(helloMsg.getIp());
					break;
					
				case STAT_GATEWAY:
				case STAT_MEMBER:
					if(helloMsg.getClusterHead()!=clusterHead)
						setStat(STAT_GATEWAY);
					break;
					
				}
				
			break;


		}

		
	/*	switch(helloMsg.getStat()){
		case STAT_CLUSTERHEAD:
			receiveHelloCH=true;
			clusterHead=helloMsg.getIp();
			log("Register to clusterhead "+clusterHead);
			sendReplyHelloMessage(clusterHead);
			setStat(STAT_MEMBER);
			break;
		case STAT_MEMBER:
			
			break;
		default:*/
			
		
	}
	
	private void addMember(NetAddress member){
		log("Add cluster member: "+member);
	}		
	
	
	  /**
	   * Sleep for a random time. 
	   * 
	   * @param time max sleep time
	   */
	  private static void randomSleep(long time)
	  {
	    JistAPI.sleep(Util.randomTime(time));
	  }	  

	  
	public void sendHelloMessage(){
		HelloMessage msg = new HelloMessage(netAddr,clusterHead,0,stat,false);
		log("Send hello");		
		NetMessage.Ip ipMsg =
	        new NetMessage.Ip(
	          msg,
	          netAddr,
	          NetAddress.ANY,
	          Constants.NET_PROTOCOL_CLUSTER,
	          Constants.NET_PRIORITY_NORMAL,
	          (byte)1);		  		
		
		self.sendIpMsg(ipMsg, MacAddress.ANY);
		

	}
	
	public void sendReplyHelloMessage(){
		randomSleep(Constants.MILLI_SECOND*100);
		HelloMessage msg = new HelloMessage(netAddr,clusterHead,0,stat,true);
		log("Send Reply hello");		
		NetMessage.Ip ipMsg =
	        new NetMessage.Ip(
	          msg,
	          netAddr,
	          NetAddress.ANY,
	          Constants.NET_PROTOCOL_CLUSTER,
	          Constants.NET_PRIORITY_NORMAL,
	          (byte)1);		  		
		
		self.sendIpMsg(ipMsg, MacAddress.ANY);
		

		
	}	
	

	public void sendIpMsg(Ip ipMsg, MacAddress destMacAddr) {
	    //transmission delay
	    randomSleep(TRANSMISSION_JITTER);
	    //send message
	    netEntity.send(ipMsg, Constants.NET_INTERFACE_DEFAULT, destMacAddr);
	}
	

	public void peek(NetMessage msg, MacAddress lastHop) {
		// TODO Auto-generated method stub

	}

	public void send(NetMessage msg) {
		// TODO Auto-generated method stub

	}

	public void receive(Message msg, NetAddress src, MacAddress lastHop,
			byte macId, NetAddress dst, byte priority, byte ttl) {
		// TODO Auto-generated method stub
		
		if(msg instanceof HelloMessage){

			HelloMessage helloMsg = (HelloMessage)msg;
			
			/*if(helloMsg.reply){
				if(stat==STAT_CLUSTERHEAD)
					receiveReplyHelloMessage(helloMsg);	
			}else*/ 
				receiveHelloMessage(helloMsg);
			
		}
	

	}


	  /**
	   * Sets network entity.
	   * 
	   * @param netEntity network entity
	   */
	  public void setNetEntity(NetInterface netEntity)
	  {
	    this.netEntity = netEntity;
	  }	
	  

	  
	  public RouteInterface.Cluster getProxy(){
		  return self;
	  }

	public NetAddress getAddress() {
		return netAddr;
	}

	public byte getStat() {
		return stat;
	}
	
	public String getStrStat(byte stat){
		switch(stat){
		case STAT_UNDECIDED:return "Undecided";
		case STAT_MEMBER:return "Member";
		case STAT_CLUSTERHEAD:return "ClusterHead";
		case STAT_GATEWAY:return "Gateway";		
		}
		return "N/A";		
	}

	public String getStrStat(){
		return getStrStat(stat);
	}


	public void setStat(byte stat) {
		if(stat!=this.stat) 
			log(" is now " + getStrStat(stat));
		
		this.stat = stat;
	}
	
	public String toString(){
		String s="";
		s+=netAddr + " is a " + getStrStat() + "\n";
		if(stat!=STAT_CLUSTERHEAD) s+="\tClusterHead : " + clusterHead + "\n" ;
		return s;
	}

}
