package jist.swans.route;

import java.util.HashSet;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.LinkedList;

import org.apache.log4j.Logger;

import jist.swans.mac.MacAddress;
import jist.swans.misc.Message;
import jist.swans.net.NetAddress;
import jist.swans.net.NetInterface;
import jist.swans.net.NetMessage;
import jist.swans.misc.Location;
import jist.swans.field.GpsInterface;
import jist.swans.loc.LocationServiceInterface;
import jist.swans.Constants;
import jist.swans.misc.Util;
import jist.runtime.JistAPI;


public class RouteMFR implements RouteInterface.Mfr {
	
	  /** The maximum amount of jitter before sending a packet. */
	  public static final long TRANSMISSION_JITTER = 1 * Constants.MILLI_SECOND;	

	  /** Time out for ReplyPos message*/
	  public static final long TIMEOUT_REPLYPOS = 100 * Constants.MILLI_SECOND;	
	  
	  /** The maximum Time-To-Live for a MFR packet. */
	  public static final byte MAX_TTL             = (byte)127;
	  
	  
	  /** Network entity. */
	  private NetInterface netEntity;
	  /** Self-referencing proxy entity. */
	  private RouteInterface.Mfr self;

	  /** local network address. */
	  private NetAddress netAddr;
	  
	  private GpsInterface gps;
	  
	  /** Hash table of location with adresse Ip key */
	  private HashSet replyPosTable;
	  
	  /** buffer for storing messages that need routes. */
	  private MessageQueue msgQueue; 
	  
	  
	  private LocationServiceInterface locService;
	  
	  private static Logger log = Logger.getLogger(RouteMFR.class.getName());	  
	

	/**
	 * Geographical location request message
	 */
	private static class RequestPosMessage implements Message {
	    /** Message size in bytes. */
	    private static final int MESSAGE_SIZE = 5;
	    
	    /**message originator IP address field. */
	    private NetAddress origIp;
	    
	    public RequestPosMessage(NetAddress origIp){
	    	this.origIp=origIp;
	    }

	    
	    public NetAddress getOrigIp() {
			return origIp;
		}
		/**
	     * Returns packet size.
	     * 
	     * @return packet size
	     */
	    public int getSize()
	    {
	      return MESSAGE_SIZE;
	    }
	    /**
	     * Store packet into byte array.
	     *
	     * @param msg destination byte array
	     * @param offset byte array starting offset
	     */
	    public void getBytes(byte[] msg, int offset)
	    {
	      throw new RuntimeException("RouteReplyMessage.getBytes() not implemented.");    
	    }	    
	    
	}
	
	/**
	 * Geographical location reply message
	 */	
	private static class ReplyPosMessage implements Message {
	    /** Message size in bytes. */
	    private static final int MESSAGE_SIZE = 9;
	    
	    /**message originator IP address field. */
	    private NetAddress origIp;
	    
	    /** Geographical position */
	    private Location loc;
	    
	    public ReplyPosMessage(NetAddress origIp,Location loc){
	    	this.origIp=origIp;
	    	this.loc=loc;
	    }

	    
	    public Location getLoc() {
			return loc;
		}



		public void setLoc(Location loc) {
			this.loc = loc;
		}



		public NetAddress getOrigIp() {
			return origIp;
		}
		/**
	     * Returns packet size.
	     * 
	     * @return packet size
	     */
	    public int getSize()
	    {
	      return MESSAGE_SIZE;
	    }
	    /**
	     * Store packet into byte array.
	     *
	     * @param msg destination byte array
	     * @param offset byte array starting offset
	     */
	    public void getBytes(byte[] msg, int offset)
	    {
	      throw new RuntimeException("RouteReplyMessage.getBytes() not implemented.");    
	    }	    
	    
	}
	
	private static class DataMessage implements Message {
	    
	    
	    /** Geographical position */
	    private Location locDest;
	    
	    /** The payload of the message. */
	    private Message content;	    
	    
	    /** node address of the next node*/
	    private NetAddress nextRecipient;
	    
	    private NetAddress dst;
	    
	    private short nextHeaderType;
	    
	    public DataMessage(Message payload,Location locDest, NetAddress nextRecipient,NetAddress dst,short nextHeaderType){
	    	this.locDest=locDest;
	    	this.content=payload;
	    	this.nextRecipient=nextRecipient;
	    	this.dst=dst;
	    	this.nextHeaderType=nextHeaderType;
	    }

	    
	    public NetAddress getNextRecipient() {
			return nextRecipient;
		}


		public void setNextRecipient(NetAddress nextRecipient) {
			this.nextRecipient = nextRecipient;
		}


		public Location getLocDest() {
			return locDest;
		}
	    
	    public Message getPayload(){
	    	return content;
	    }


		/**
	     * Returns packet size.
	     * 
	     * @return packet size
	     */
	    public int getSize()
	    {
	      return content.getSize()+14;
	    }
	    /**
	     * Store packet into byte array.
	     *
	     * @param msg destination byte array
	     * @param offset byte array starting offset
	     */
	    public void getBytes(byte[] msg, int offset)
	    {
	      throw new RuntimeException("RouteReplyMessage.getBytes() not implemented.");    
	    }


		public NetAddress getDst() {
			return dst;
		}


		public short getNextHeaderType() {
			return nextHeaderType;
		}	    
	    
	}

	  /**
	   * A MessageQueue object temporarily stores transport-layer messages while routes are
	   * being determined.  When route information becomes available, the messages are then
	   * sent along the routes.
	   */
	  private static class MessageQueue
	  {
	    /** list of IP messages (with type NetMessage.Ip). */
	    private LinkedList list;
	    /** reference to this RouteAodv instance. */
	    private RouteMFR thisNode; 
	    
	    /**
	     * Constructs a MessageQueue object, with an empty list.
	     * 
	     * @param thisNode reference to this RouteAodv instance
	     */
	    public MessageQueue(RouteMFR thisNode)
	    {
	      list = new LinkedList();
	      this.thisNode = thisNode;
	    }
	    
	    /**
	     * Adds a NetMessage.Ip to the queue.
	     * 
	     * @param msg message to add to queue
	     */
	    public void add(NetMessage.Ip msg)
	    {
	      list.addLast(msg);
	    }
	    
	    /**
	     * Sends all messages in queue destined for a given destination via a given next hop.
	     * 
	     * @param dest destination address
	     * @param nextHop next hop address
	     */
	    public void dequeueAndSend()
	    {
	      for (int i=0; i<list.size(); i++)
	      {

	    	  NetMessage.Ip msg = (NetMessage.Ip)list.remove(i);
	    	  
	    	  DataMessage mfrMsg = (DataMessage )msg.getPayload();
	    	  
	    	  //compute the distance between me and destination
	    	  float remainingDistance = mfrMsg.getLocDest().distance(thisNode.gps.getPos()); 	    	  
	    	  float minDistance = Float.MAX_VALUE;
	    	  
	    	  Iterator it =  thisNode.replyPosTable.iterator();
	    	 
	    	  
	    	  while(it.hasNext()){
	    		  
	    		  ReplyPosMessage repPosMsg = (ReplyPosMessage) it.next();
	    		  
	    		  //compute the distance between the destination and the candidate node
	    		  float dist = mfrMsg.getLocDest().distance(repPosMsg.getLoc());
	    		  
	    		  if(dist<remainingDistance && (dist<minDistance)){
	    			  minDistance=dist;
	    			  mfrMsg.setNextRecipient(repPosMsg.getOrigIp()) ;
	    		  }
	    		  
	    		  
	    	  }
	    	  

	    	  if(mfrMsg.getNextRecipient()!=null){
	    		  NetMessage.Ip newMsg = new NetMessage.Ip(	
	    				  msg.getPayload(),
	    		          msg.getSrc(),
	    		          mfrMsg.getNextRecipient(),
	    		          msg.getProtocol(),
	    		          msg.getPriority(),
	    		          msg.getTTL());		  

	    		  thisNode.self.sendIpMsg(newMsg, MacAddress.ANY);
	    	  }
	      }
	    }
	    
	    /**
	     * Removes all messages bound for a given destination.
	     * 
	     * @param dest destination net address
	     */
	  /*  public void removeMsgsForDest(NetAddress dest)
	    {
	      Iterator itr = list.listIterator(0);
	      while (itr.hasNext())
	      {
	        NetMessage.Ip msg = (NetMessage.Ip)itr.next();
	        if (msg.getDst().equals(dest))
	        {
	          itr.remove();
	        }
	      }
	    }*/
	  }
	  
	  
	  public RouteMFR(NetAddress netAddr)
	  {
		    // proxy entity
		    this.self = (RouteInterface.Mfr)JistAPI.proxy(this, RouteInterface.Mfr.class);

		    
		    //instantiate message queue
		    this.msgQueue = new MessageQueue(this);	
		    this.netAddr = netAddr;
		    this.replyPosTable = new HashSet();
	  }
	  
	  /**
	   * Send a reply position message 
	   *
	   */
	  private void sendReplyPos(NetAddress dst, MacAddress lastMac){
		  
		  Location loc = gps.getPos();
		  ReplyPosMessage	repPosMsg = new ReplyPosMessage(netAddr,loc);	  
		  
	      NetMessage.Ip repPosMsgIp =
		        new NetMessage.Ip(
		          repPosMsg,
		          netAddr,
		          dst,
		          Constants.NET_PROTOCOL_MFR,
		          Constants.NET_PRIORITY_NORMAL,
		          (byte)1);		  
		  
		  self.sendIpMsg(repPosMsgIp,lastMac);	
	  }
	  
	  /**
	   * broadcast a Request Pos Message
	   */
	  private void broadcastRequestPos(){
	      	      
	      // create RREQ message
	      RequestPosMessage reqMsg =  new RequestPosMessage(netAddr);
	      
	      // create IP message containing the RREQ message
	      NetMessage.Ip rreqMsgIp =
	        new NetMessage.Ip(
	        	reqMsg,
	          netAddr,
	          NetAddress.ANY,
	          Constants.NET_PROTOCOL_MFR,
	          Constants.NET_PRIORITY_NORMAL,
	          (byte)1);
	      
	      self.sendIpMsg(rreqMsgIp, MacAddress.ANY);		  
	  }
	  
	  public void RequestPosTimeOut(){
			//envoye tous les messages de la file d'attente
			msgQueue.dequeueAndSend();	
			replyPosTable.clear();
	  }
	  
	  private void forwardDataMessage(DataMessage dataMsg,NetAddress  dest,byte ttl){
		  
		  	dataMsg.nextRecipient=null;
		  
		    NetMessage.Ip msgIp =
		        new NetMessage.Ip(
		          dataMsg,
		          netAddr,
		          dest,
		          Constants.NET_PROTOCOL_MFR,
		          Constants.NET_PRIORITY_NORMAL,
		          (byte)(ttl - 1) );			    

			//add the packet to the queue
			msgQueue.add(msgIp);
			
			//cherche la position des voisins
			broadcastRequestPos();
			
			//attend un peu
			JistAPI.sleep(TIMEOUT_REPLYPOS);
			
			self.RequestPosTimeOut();
	  }
	  
	  /**
	   * Sleep for a random time. 
	   * 
	   * @param time max sleep time
	   */
	  private static void randomSleep(long time)
	  {
	    JistAPI.sleep(Util.randomTime(time));
	  }	  
	  /**
	   * Sets network entity.
	   * 
	   * @param netEntity network entity
	   */
	  public void setNetEntity(NetInterface netEntity)
	  {
	    this.netEntity = netEntity;
	  }	
	  
	  public void setGpsEntity(GpsInterface gps){
		  this.gps=gps;
	  }
	  
	  public RouteInterface.Mfr getProxy(){
		  return self;
	  }
	
	public void peek(NetMessage msg, MacAddress lastHop) {
		// TODO Auto-generated method stub
	}

	public void send(NetMessage msg) {

	    NetMessage.Ip ipMsg = (NetMessage.Ip)msg;
	    if (ipMsg.getDst().equals(netAddr))
	    {
	      throw new RuntimeException("Message is already at destination.  Why is RouteAodv.send() being called?");
	    }		
	    
	    if (ipMsg.getProtocol() == Constants.NET_PROTOCOL_MFR)
	    {
	      // We're not really supposed to "send" this packet 
	    }
	    else
	    {
	    
		    //broadcastRequestPos();
		    
		    Location locDest=locService.getLocation(ipMsg.getDst());
		    
		    DataMessage dataMsg = new DataMessage(ipMsg.getPayload(),locDest,null,ipMsg.getDst(),ipMsg.getProtocol());
		      
		    forwardDataMessage(dataMsg,ipMsg.getDst(),MAX_TTL);		
	    
	    }
	}

	public void receive(Message msg, NetAddress src, MacAddress lastHop,
			byte macId, NetAddress dst, byte priority, byte ttl) {
		// TODO Auto-generated method stub
		
		
	    /*if (msg instanceof NetMessage.Ip)
	    {
	      NetMessage.Ip ipMsg = (NetMessage.Ip)msg;
	      
	      if (ipMsg.getProtocol() == Constants.NET_PROTOCOL_MFR)
	      {
	    	Message mfrMsg = ipMsg.getPayload();
			*/
		
			//TODO attention pas de instanceof ?
			if(msg instanceof RequestPosMessage){
				if(log.isDebugEnabled())
					log.debug("Request message " + netAddr );
				RequestPosMessage reqPosMsg = (RequestPosMessage)msg;
				sendReplyPos(reqPosMsg.origIp, lastHop);
			}
			
			if(msg instanceof ReplyPosMessage){
				ReplyPosMessage repPosMsg = (ReplyPosMessage)msg;				
				
				if(log.isDebugEnabled())
					log.debug("Reply message (" + netAddr + " " + repPosMsg.getLoc() +  ")");		

				
				replyPosTable.add(repPosMsg);
			}	
			
			if(msg instanceof DataMessage){
				DataMessage dataMsg = (DataMessage)msg;
				if(log.isDebugEnabled())
					log.debug("Data message " + netAddr);	
				
				if(dataMsg.getDst().equals(netAddr)){
					
				    NetMessage.Ip newIp = new NetMessage.Ip(dataMsg.getPayload(), src, dst,
				    		dataMsg.getNextHeaderType(), priority, ttl);     
					
					netEntity.receive(newIp, lastHop, macId, false);
					
					
				}else
				
				//if I am the next hop, forward
				if (dataMsg.getNextRecipient().equals(netAddr))
					forwardDataMessage(dataMsg,dataMsg.getNextRecipient(),ttl);
			}
			
	     /* }
	      
	    }*/
		
	}
	  /**
	   * Sends IP message after transmission delay.
	   * 
	   * @param ipMsg IP message to send
	   * @param destMacAddr next hop mac address
	   */
	  public void sendIpMsg(NetMessage.Ip ipMsg, MacAddress destMacAddr)
	  {
	    //transmission delay
	    randomSleep(TRANSMISSION_JITTER);
	    //send message
	    netEntity.send(ipMsg, Constants.NET_INTERFACE_DEFAULT, destMacAddr);
	    
	    System.out.println("Send a packet : "+ipMsg);
	  }

	public void setLocService(LocationServiceInterface locService) {
		this.locService = locService;
	}
	
}
